@@ -13,7 +13,7 @@ use tcgeneric::*;
 use crate::chain::{self, Chain, ChainType, Schema};
 use crate::fs;
 use crate::object::{InstanceClass, InstanceExt};
-use crate::scalar::{Link, LinkHost, OpRef, Scalar, Value};
+use crate::scalar::{Link, LinkHost, OpRef, Refer, Scalar, Value};
 use crate::txn::{Actor, Txn, TxnId};
 
 use super::Cluster;
@@ -58,6 +58,27 @@ pub async fn instantiate(
                 }
             }
             Scalar::Op(op_def) => {
+                let op_def = if op_def.is_write() {
+                    // make sure not to replicate ops internal to this OpDef
+                    let op_def = op_def.reference_self(link.path());
+
+                    // make sure not to duplicate requests to other clusters
+                    for (id, provider) in op_def.form() {
+                        if provider.is_inter_service_write(link.path()) {
+                            return Err(TCError::unsupported(format!(
+                                "replicated op {} may not perform inter-service writes: {}",
+                                id, provider
+                            )));
+                        }
+                    }
+
+                    op_def
+                } else {
+                    // make sure to replicate all write ops internal to this OpDef
+                    // by routing them through the kernel
+                    op_def.dereference_self(link.path())
+                };
+
                 cluster_proto.insert(id, Scalar::Op(op_def));
             }
             other => {
@@ -76,6 +97,8 @@ pub async fn instantiate(
 
     let mut chains = Map::<Chain>::new();
     for (id, (class, schema)) in chain_schema.into_iter() {
+        debug!("load chain {} of type {} with schema {}", id, class, schema);
+
         let dir = dir.get_or_create_dir(txn_id, id.clone()).await?;
         let chain = chain::load(txn, class, schema, dir).await?;
         chains.insert(id, chain);
