@@ -8,7 +8,7 @@ use std::ops::Deref;
 use std::sync::Arc;
 
 use async_trait::async_trait;
-use futures::future::{join_all, try_join_all};
+use futures::future::{join_all, try_join_all, Future, FutureExt};
 use futures::{join, StreamExt};
 use log::{debug, info, warn};
 use safecast::TryCastFrom;
@@ -25,14 +25,14 @@ use crate::scalar::{Link, OpDef, Value};
 use crate::state::State;
 use crate::txn::{Actor, Scope, Txn, TxnId};
 
-mod load;
-mod owner;
-
 use owner::Owner;
 
 use futures::stream::FuturesUnordered;
 pub use load::instantiate;
 
+mod load;
+mod owner;
+
 /// The name of the endpoint which serves a [`Link`] to each of this [`Cluster`]'s replicas.
 pub const REPLICAS: Label = label("replicas");
 
@@ -98,15 +98,7 @@ impl Cluster {
 
     /// Claim ownership of the given [`Txn`].
     pub async fn claim(&self, txn: &Txn) -> TCResult<Txn> {
-        let last_commit = self.confirmed.read().await;
-        if txn.id() <= &*last_commit {
-            return Err(TCError::unsupported(format!(
-                "cluster at {} cannot claim transaction {} because the last commit is at {}",
-                self.link,
-                txn.id(),
-                *last_commit
-            )));
-        }
+        self.validate_txn_id(txn.id()).await?;
 
         let mut owned = self.owned.write().await;
         if owned.contains_key(txn.id()) {
@@ -169,7 +161,7 @@ impl Cluster {
             .grant(&self.actor, self.link.path().clone(), vec![scope])
             .await?;
 
-        OpDef::call(op.into_def(), txn, context).await
+        OpDef::call(op.into_form(), txn, context).await
     }
 
     /// Trust the `Cluster` at the given [`Link`] to issue the given auth [`Scope`]s.
@@ -195,6 +187,12 @@ impl Cluster {
         Ok(())
     }
 
+    /// Claim leadership of the given [`Txn`].
+    pub async fn lead(&self, txn: Txn) -> TCResult<Txn> {
+        self.validate_txn_id(txn.id()).await?;
+        txn.lead(&self.actor, self.link.path().clone()).await
+    }
+
     /// Add a replica to this cluster.
     pub async fn add_replica(&self, txn: &Txn, replica: Link) -> TCResult<()> {
         let self_link = txn.link(self.link.path().clone());
@@ -265,6 +263,18 @@ impl Cluster {
         Ok(())
     }
 
+    async fn validate_txn_id(&self, txn_id: &TxnId) -> TCResult<()> {
+        let last_commit = self.confirmed.read().await;
+        if txn_id <= &*last_commit {
+            Err(TCError::unsupported(format!(
+                "cluster at {} cannot claim transaction {} because the last commit is at {}",
+                self.link, txn_id, *last_commit
+            )))
+        } else {
+            Ok(())
+        }
+    }
+
     async fn replicate(&self, txn: &Txn) -> TCResult<()> {
         let replication = self.chains.iter().map(|(name, chain)| {
             let mut path = self.link.path().to_vec();
@@ -278,12 +288,66 @@ impl Cluster {
         Ok(())
     }
 
+    pub async fn replicate_write<F: Future<Output = TCResult<()>>, W: Fn(Link) -> F>(
+        &self,
+        txn: Txn,
+        write: W,
+    ) -> TCResult<()> {
+        let mut replicas = self.replicas(txn.id()).await?;
+        replicas.remove(&txn.link(self.link().path().clone()));
+        debug!("replicating write to {} replicas", replicas.len());
+
+        let max_failures = replicas.len() / 2;
+        let mut failed = HashSet::with_capacity(replicas.len());
+        let mut succeeded = HashSet::with_capacity(replicas.len());
+
+        {
+            let mut results = FuturesUnordered::from_iter(
+                replicas
+                    .into_iter()
+                    .map(|link| write(link.clone()).map(|result| (link, result))),
+            );
+
+            while let Some((replica, result)) = results.next().await {
+                match result {
+                    Err(cause) if cause.code() == ErrorType::Conflict => return Err(cause),
+                    Err(ref cause) => {
+                        debug!("replica at {} failed: {}", replica, cause);
+                        failed.insert(replica);
+                    }
+                    Ok(()) => {
+                        debug!("replica at {} succeeded", replica);
+                        succeeded.insert(replica);
+                    }
+                };
+
+                if failed.len() > max_failures {
+                    assert!(result.is_err());
+                    return result;
+                }
+            }
+        }
+
+        if !failed.is_empty() {
+            let failed = Value::from_iter(failed);
+            try_join_all(
+                succeeded
+                    .into_iter()
+                    .map(|replica| txn.delete(replica.append(REPLICAS.into()), failed.clone())),
+            )
+            .await?;
+        }
+
+        Ok(())
+    }
+
     pub async fn mutate(&self, txn: &Txn, participant: Link) -> TCResult<()> {
         if participant.path() == self.link.path() {
             log::warn!(
                 "got participant message within Cluster {}",
                 self.link.path()
             );
+
             return Ok(());
         }
 
