@@ -11,7 +11,7 @@ use destream::de::{self, Decoder, FromStream};
 use destream::en::{EncodeMap, Encoder, IntoStream, ToStream};
 use futures::{try_join, TryFutureExt};
 use log::debug;
-use safecast::{Match, TryCastFrom, TryCastInto};
+use safecast::{TryCastFrom, TryCastInto};
 
 use tc_error::*;
 use tcgeneric::*;
@@ -82,6 +82,33 @@ pub enum Subject {
 }
 
 impl Subject {
+    fn dereference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::Ref(id_ref, suffix) if id_ref.id() == &SELF => {
+                let mut path = path.clone();
+                path.extend(suffix);
+                Self::Link(path.into())
+            }
+            other => other,
+        }
+    }
+
+    fn is_self(&self) -> bool {
+        match self {
+            Self::Ref(id_ref, _) => id_ref.id() == &SELF,
+            _ => false,
+        }
+    }
+
+    fn reference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::Link(link) if link.path().starts_with(path) => {
+                Self::Ref(IdRef::from(SELF), link.path()[path.len()..].to_vec().into())
+            }
+            other => other,
+        }
+    }
+
     fn requires(&self, deps: &mut HashSet<Id>) {
         match self {
             Self::Ref(id_ref, _) if id_ref.id() != &SELF => id_ref.requires(deps),
@@ -276,6 +303,56 @@ impl Instance for OpRef {
 
 #[async_trait]
 impl Refer for OpRef {
+    fn dereference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::Get((subject, key)) if subject.is_self() => {
+                Self::Get((subject.dereference_self(path), key))
+            }
+            Self::Put((subject, key, value)) if subject.is_self() => {
+                Self::Put((subject.dereference_self(path), key, value))
+            }
+            Self::Post((subject, params)) if subject.is_self() => {
+                Self::Post((subject.dereference_self(path), params))
+            }
+            Self::Delete((subject, key)) if subject.is_self() => {
+                Self::Delete((subject.dereference_self(path), key))
+            }
+            op_ref => op_ref,
+        }
+    }
+
+    fn is_inter_service_write(&self, cluster_path: &[PathSegment]) -> bool {
+        let subject = match self {
+            Self::Put((Subject::Link(link), _, _)) => Some(link),
+            Self::Delete((Subject::Link(link), _)) => Some(link),
+            _ => None,
+        };
+
+        if let Some(link) = subject {
+            !link.path().starts_with(cluster_path)
+        } else {
+            false
+        }
+    }
+
+    fn reference_self(self, path: &TCPathBuf) -> Self {
+        match self {
+            Self::Get((subject, key)) if subject.is_self() => {
+                Self::Get((subject.reference_self(path), key))
+            }
+            Self::Put((subject, key, value)) if subject.is_self() => {
+                Self::Put((subject.reference_self(path), key, value))
+            }
+            Self::Post((subject, params)) if subject.is_self() => {
+                Self::Post((subject.reference_self(path), params))
+            }
+            Self::Delete((subject, key)) if subject.is_self() => {
+                Self::Delete((subject.reference_self(path), key))
+            }
+            op_ref => op_ref,
+        }
+    }
+
     fn requires(&self, deps: &mut HashSet<Id>) {
         match self {
             Self::Get((subject, key)) => {
@@ -399,15 +476,16 @@ impl OpRefVisitor {
     pub fn visit_ref_value<E: de::Error>(subject: Subject, params: Scalar) -> Result<OpRef, E> {
         match params {
             Scalar::Map(params) => Ok(OpRef::Post((subject, params))),
-            Scalar::Tuple(params) if params.matches::<(Scalar, Scalar)>() => {
-                let (key, value) = params.opt_cast_into().unwrap();
-                Ok(OpRef::Put((subject, key, value)))
-            }
-            Scalar::Tuple(params) if params.matches::<(Scalar,)>() => {
-                let (key,) = params.opt_cast_into().unwrap();
+            Scalar::Tuple(mut tuple) if tuple.len() == 1 => {
+                let key = tuple.pop().unwrap();
                 Ok(OpRef::Get((subject, key)))
             }
-            other => Err(de::Error::invalid_type(other, &"OpRef")),
+            Scalar::Tuple(mut tuple) if tuple.len() == 2 => {
+                let value = tuple.pop().unwrap();
+                let key = tuple.pop().unwrap();
+                Ok(OpRef::Put((subject, key, value)))
+            }
+            other => Err(de::Error::invalid_value(other, "OpRef parameters")),
         }
     }
 }
@@ -453,7 +531,7 @@ impl<'en> ToStream<'en> for OpRef {
         let mut map = e.encode_map(Some(1))?;
 
         match self {
-            OpRef::Get((path, key)) => map.encode_entry(path.to_string(), key)?,
+            OpRef::Get((path, key)) => map.encode_entry(path.to_string(), (key,))?,
             OpRef::Put((path, key, value)) => map.encode_entry(path.to_string(), (key, value))?,
             OpRef::Post((path, data)) => map.encode_entry(path.to_string(), data.deref())?,
             OpRef::Delete((path, key)) => {
@@ -484,6 +562,12 @@ impl<'en> IntoStream<'en> for OpRef {
     }
 }
 
+impl fmt::Debug for OpRef {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        fmt::Display::fmt(self, f)
+    }
+}
+
 impl fmt::Display for OpRef {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         let class = self.class();
